import { useRerender } from "@hmans/use-rerender"
import { extend, InstancedMeshProps } from "@react-three/fiber"
import { getShaderRootForMaterial } from "material-composer-r3f"
import React, {
  createContext,
  forwardRef,
  Ref,
  useContext,
  useImperativeHandle,
  useLayoutEffect,
  useRef
} from "react"
import { Material } from "three"
import { Particles as ParticlesImpl } from "vfx-composer"
import { useFrameEffect } from "./lib/useFrameEffect"

export type ParticlesProps = Omit<
  InstancedMeshProps,
  "material" | "args" | "ref"
> & {
  ref?: Ref<ParticlesImpl>
  args?: ConstructorParameters<typeof ParticlesImpl>
  material?: Material
  capacity?: number
  safetyCapacity?: number
}

extend({ VfxComposerParticles: ParticlesImpl })

declare global {
  namespace JSX {
    interface IntrinsicElements {
      vfxComposerParticles: ParticlesProps
    }
  }
}

const Context = createContext<ParticlesImpl | null>(null)

export const useParticlesContext = () => useContext(Context)

export const Particles = forwardRef<ParticlesImpl, ParticlesProps>(
  (
    { children, capacity, safetyCapacity, geometry, material, ...props },
    ref
  ) => {
    const rerender = useRerender()
    const particles = useRef<ParticlesImpl>(null!)

    /*
    We need to be able to react to the particle mesh's material changing. Currently,
    the best way of doing this is to use a custom per-frame callback that will monitor
    the material for changes, and re-initialize the particles mesh when it does.
    We will find a cleaner way of doing this some time in the future.
    */

    useFrameEffect(
      () => particles.current.material as Material,
      (material) => {
        if (!material) return
        const root = getShaderRootForMaterial(material)
        if (root) particles.current.setupParticles(root)
      },
      -100
    )

    /*
    We need to re-render this beautiful component once, because it's highly
    likely that the material driving the particle effect is only declared
    within its children.
    */
    useLayoutEffect(() => {
      rerender()
    }, [particles])

    useImperativeHandle(ref, () => particles.current)

    return (
      <vfxComposerParticles
        args={[geometry, material, capacity, safetyCapacity]}
        ref={particles}
        {...props}
      >
        <Context.Provider value={particles.current}>
          {children}
        </Context.Provider>
      </vfxComposerParticles>
    )
  }
)
